:- module( debug_call, [ 
                           debug_call/2,
					  debug_call/3,
					  debug_call/4,
					  debug_chain/2, debug_chain/3,
					  debug_on/1,
                           debug_portray/2,
					  debug_reset/2,
                           debugging_status/2,
					  debug_topic/2,
					  debug_call_version/2
                       ] ).

/** <module> Debugging with calls.

Avoids running goals to produce output that is only
relevant while debugging. Includes pre-canned often used calls.

==

?- debug( ex ).
?- debug_call( ex, length, '', list1/[x,y,z] ).
% Length for list, list1: 3

?- debug_call( ex, length, 'some prefix', [list1,list2]/[[x,y,z],[a,b,c]] ).
% some prefix lengths for lists, list1: 3, list2: 3

?- debug_call( ex, dims, [m1,m2]/[[a(x),a(y),a(z)],[xy(a,b),xy(c,d),xy(e,f)]] ).
%  Dimensions for matrices,  (m1) nR: 3, nC: 1. (m2) nR: 3, nC: 2.

?- debug_call( ex, wrote, loc(file,csv) ).
% Could not locate wrote on file specified by: file, and extensions: csv
?- csv_write_file( 'file.csv', [] ).

?- debug_call( ex, wrote, loc(file,csv) ).
% Wrote on file: '/home/nicos/pl/lib/src/trace/file.csv'

?- debug_call( ex, task(stop), 'write on file' ).
At 15:44:1 on 2nd of Jul 2014 finished task: write on file.

==

@author nicos angelopoulos
@see http://stoics.org.uk/~nicos/sware/debug_call/
@tbd options_debug( Opts, Mess, Args )  only writes if Opts contains debug(true). maybe this should be part of pack(options)

*/

/** debug_call_version( -Version, -Date ).

Current version and release date for the library.

*/
debug_call_version( 0:1:0, date(2016,3,5) ).

% :- use_module( library(requires) ). % it doesn't in the public version


%% debug_call( +Topic, +Goal ).
%
%  Only call debug if we are debugging Topic.
%
% propose this to be included to debug.pl
%
debug_call( Topic, Goal ) :-
	debugging( Topic ),
	!,
	call( Goal ).
debug_call( _Topic, _Goal ).

%% debug_chain( +TopicCond, +TopicDep ).
%% debug_chain( +TopicCond, +TopicDep, -TDprior ).
%
% If already debugging TopicCond, then also start debugging TopicDep ).
% TDprior is true if TopicDep was already debugging, else is false.
% Current implementation sets TDprior to true whenever Topic is not
% debugged, as it assumes that this value best suit independent fluctuation
% of TopicDep. Only in the case of debug_chain/2, TopicDep can be a list.
%
% @author nicos angelopoulos
% @version  0.1 2014/4/4
% @see debug_reset/2
%
debug_chain( Topic, Then ) :-
	to_list( Then, Thens ),
	maplist( debug_chain(Topic), Thens, _Priors ).

debug_chain( Topic, Then, Prior ) :-
	debugging( Topic ),
	!,
	debugging_status( Then, Prior ),
	debug( Then ).
debug_chain( _Topic, _Then, true ). 
	% setting 3rd to true is a bit presumptious of its uses later on

%% debugging_status( +Topic, -Status ).
%
% Status == true iff debugging(Topic) succeeds. Else, it is false.
% Similar to debugging/2, but does not fail for undefined Topic.
%==
% ?- debug( something ).
% true.
% ?- debugging_status( something, Some ).
% Some = true.
% ?- debugging_status( some_else, Else ).
% Else = false.
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/7/23
debugging_status( Topic, Status ) :-
	debugging( Topic ),
	!,
	Status = true.
debugging_status( _Topic, false ).

%% debug_reset( +Prior, +Topic ).
%
% Reset Topic according to Prior: true does nothing, false turns Topic off.
%
%==
% ?- nodebug( chained ).
% true.
% ?- debug( testo ).
% Warning: testo: no matching debug topic (yet)
% true.
% ?- debug( chained, 'debugs chains 1', [] ).
% true.
% ?- debug_chain( testo, chained, Prior ).
% Prior = false.
% ?- debug( chained, 'debugs chains 2', [] ).
% % debugs chains 2
% true.
% ?- Prior = false, debug_reset( Prior, chained ).
% Prior = false.
% ?- debug( chained, 'debugs chains 3', [] ).
% true
%==
% @author nicos angelopoulos
% @version  0.1 2014/7/23
% @see debug_chain/3
debug_reset( false, Topic ) :-
	nodebug( Topic ).
debug_reset( true, _Topic ).

%% debug_topic( +Flag, +Topic ).
%
% Start debugging Topic if Flag == true.
% 
%==
% ?- debug_topic( true, example ).
%==
% @author nicos angelopoulos
% @version  0.1 2014/12/10
% @see options_append/4
%
debug_topic( true, Topic ) :-
	!,
	debug( Topic ).
debug_topic( _, _Topic ).

%% debug_on( +Topic ).
%
% As debug/1, but do not print warning if topic is not known.
%
debug_on( Topic ) :-
 	asserta( prolog_debug:debugging(Topic,true,[user_error])).

%% debug_portray( +Topic, +Term ).
%
%  Call portray_clause(Term) if we are debugging Topic.
%
% @author nicos angelopoulos
% @version  0.1
%
debug_portray( Topic, Term ) :-
	debugging( Topic ),
	!,
	portray_clause( Term ).
debug_portray( _Topic, _Term ).

%% debug_call( +Topic, +Goal, +Args ).
%% debug_call( +Topic, +Goal, +Mess, +Args ).
%
%  Automates often used debug calls. When Mess is missing it is assumed to be ''.
%
%  Goal in:
%   * *length* prints the lenghts of a bunch of lists. Args should be ListNames/Lists. 
%   uses non list ListNames if debuging the length of a single list, in which case
%   message in the singular is used.
%   * *dims* prints the dimensions of matrix, see mtx_dims/3
%
%   * *wrote* reports the writting of output on a file. Arg should be file specification suitable for locate/3.
%   Either loc(File,Exts) or simply File in which case Exts = ''.
% 
%   * *read* reports reading from a file. Arg should be file specification suitable for locate/3.
%   Either loc(File,Exts) or simply File in which case Exts = ''.
%
%   * task(Wch)  time of start/stop of a task. Other values are allowed put printed as is. The known values
%   translate to starting/finshed.
% 
%   * call(Goal)  call Goal before printing debugging message.
%
% ==
%   ?- debug( ex ).
%   ?- debug_call( ex, length, '', list1/[x,y,z] ).
%   % Length for list, list1: 3
%  
%   ?- debug_call( ex, length, 'some prefix', [list1,list2]/[[x,y,z],[a,b,c]] ).
%   % some prefix lengths for lists, list1: 3, list2: 3
%      
%   ?- debug_call( ex, wrote, loc(file,csv) ).
%   % Could not locate wrote on file specified by: file, and extensions: csv
%   ?- csv_write_file( 'file.csv', [] ).
% 
%   ?- debug_call( ex, wrote, loc(file,csv) ).
%   % Wrote on file: '/home/nicos/pl/lib/src/trace/file.csv'
% 
%    ?- debug_call( ex, task(stop), 'write on file' )
%    % At 15:44:1 on 2nd of Jul 2014 finished task: write on file.
%    
%    ?- debug_call( ex, (length([a,b,c],L),write(len(L)),nl) ).
%    len(3)
%    L = 3.
% ==
%
% @author nicos angelopoulos
% @version  0.1 2014/03/27
% @version  0.2 2014/04/24  added wrote
% @version  0.3 2014/07/2   added task
% @version  0.4 2014/09/22  added renamed from debug_call/3
%
debug_call( Topic, Goal, Args ) :-
	debug_call( Topic, Goal, '', Args ).

debug_call( Topic, Goal, Mess, Args ) :-
	debugging( Topic ),
	!,
	debugging_call( Topic, Goal, Mess, Args ).
debug_call( _Topic, _Goal, _Mess, _Args ).

debugging_call( Topic, Goal, Mess, Args ) :-
	debug_call_topic( Goal, Mess, Args, Topic ),
	!.
debugging_call( Topic, call(Goal), Mess, Args ) :-
	call( Goal ),
	!,
	debug( Topic, Mess, Args ).

debugging_call( Topic, Goal, Mess, Args ) :-
	Called = debug_call(Topic,Goal,Mess,Args),
	message_report( 'failure ignored on: ~w', Called, warning ).

%% debug_consec( +Topic, +Mess, +Args ).
%% debug_consec( +Topic, +Clrs, +Mess, +Args ).
% 
% Alternate the colours of printing messages on Topic,
% from those in Clrs. When missing these are [blue,magenta].
% 
%==
% ?- debug( dbg ).
% ?- debug_consec( dbg, 'what:~w', when' ).
% % what: where   %  % in blue
% ?- debug_consec( dbg, 'what:~w', when' ).
% % what: where   %  % in magenta
% ?- debug_consec( dbg, 'what:~w', when' ).
% % what: where   %  % in blue
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/7/24
%
debug_consec( Topic, Mess, Args ) :-
	Clrs = [blue,magenta],
	debug_consec( Topic, Clrs, Mess, Args ).

debug_consec( Topic, Clrs, Mess, Args ) :-
	debugging( Topic ),
	!,
	debug_consec_topic( Topic, Clrs, Mess, Args ).
debug_consec( _Topic, _Clrs, _Mess, _Args ).

debug_consec_topic( Topic, Clrs, Mess, Args ) :-
	with_output_to( atom(Topicat), write_term(Topic,[]) ),
	( nb_current(Topicat,Value) -> true; Value = 1 ),
	nth1( Value, Clrs, Clr ),
	debug_consec_color( Topic, Clr, Mess, Args ),
	length( Clrs, Len ),
	( Value < Len -> Next is Value + 1; Next is 1 ),
	nb_setval( Topicat, Next ).

debug_consec_color( Topic, Clr, Mess, Args ) :-
	user:message_property( debug(_), color(Attrs) ),
	!,
	retractall( user:message_property(debug(_),color(_)) ),
	assert( user:message_property(debug(_),color(fg(Clr))) ),
	debug( Topic, Mess, Args ),
	retractall( user:message_property(debug(_),color(_)) ),
	assert( user:message_property(debug(_),color(Attrs)) ).
debug_consec_color( Topic, Clr, Mess, Args ) :-
	assert( user:message_property(debug(_),color(fg(Clr))) ),
	debug( Topic, Mess, Args ),
	retractall( user:message_property(debug(_),color(_)) ).

debug_call_topic( length, Pfx, NamesPrv/ListsPrv, Topic ) :-
							% add version without names
	( is_list(NamesPrv) -> Names=NamesPrv, ListsPrv=Lists, With = 'Lengths for lists, '
	                       ; [NamesPrv] = Names, [ListsPrv]=Lists, With = 'Length for list, ' 
	),
	debug_message_prefixed( Pfx, With, Prefixed ),
	maplist( length, Lists, Lengths ),
	Sep = ': ~w, ',
	atomic_list_concat( Names, Sep, Right ),
	atom_concat( Right, ': ~w', Rightmost ),
	atom_concat( Prefixed, Rightmost, Message ),
	debug( Topic, Message, Lengths ). % do the messaging !
debug_call_topic( dims, Pfx, NamesPrv/MtxsPrv, Topic ) :-
	( is_list(NamesPrv) -> Names=NamesPrv, MtxsPrv=Mtxs, With = 'Dimensions for matrices, '
	                       ; [NamesPrv] = Names, [MtxsPrv]=Mtxs, With = 'Dimensions for matrix, ' 
	),
	debug_message_prefixed( Pfx, With, Prefixed ),
	maplist( debug_mtx_dims, Mtxs, NRows, NCols ),

	findall( PartM, (member(_,Names),PartM=' (~w) nR: ~d, nC: ~d.'), MParts ),
	atomic_list_concat( MParts, '', Right ),
	findall( [Name,NRow,NCol], (nth1(N,Names,Name),nth1(N,NRows,NRow),nth1(N,NCols,NCol)), NNest ),
	flatten( NNest, Vars ),
	% write( debug(Topic, Message, Dims ) ), nl,
	atom_concat( Prefixed, Right, Message ),
	debug( Topic, Message, Vars ). % do the messaging !

debug_call_topic( wrote, Pfx, ForLoc, Topic ) :-
	( ForLoc = loc(Spec,Ext) -> true; Spec=ForLoc, Ext = '' ),
	catch( locate(Spec,Ext,Loc), Excp, true ),
	MessW = 'Wrote on file: ~p',
	debug_call_location_exception_message( Excp, Loc, MessW, Mess, Args ),
	debug_message_prefixed( Pfx, Mess, Prefixed ),
	debug( Topic, Prefixed, Args ).
debug_call_topic( read, Pfx, ForLoc, Topic ) :-
	( ForLoc = loc(Spec,Ext) -> true; Spec=ForLoc, Ext = '' ),
	catch( locate(Spec,Ext,Loc), Excp, true ),
	MessW = 'Read from file: ~p',
	debug_call_location_exception_message( Excp, Loc, MessW, Mess, Args ),
	debug_message_prefixed( Pfx, Mess, Prefixed ),
	debug( Topic, Prefixed, Args ).

debug_call_topic( task(Whc), Pfx, Task, Topic ) :-
	datime_to_readable( Readable ),
	debug_call_topic_time_which_readable( Whc, Whcable ),
	atomic_list_concat( [Readable,' ',Whcable,' task: ', Task,'.'], Mess ),
	debug_message_prefixed( Pfx, Mess, Prefixed ),
	debug( Topic, Prefixed, [] ).

debug_call_topic_time_which_readable( Wch, Wchable ) :-
	debug_call_topic_time_which_readable_known( Wch, Wchable ),
	!.
debug_call_topic_time_which_readable( Wch, Wch ).

debug_call_topic_time_which_readable_known( start, starting ).
debug_call_topic_time_which_readable_known( finish, finished ).

debug_call_location_exception_message( Var, Loc, MessI, MessO, Args ) :-
	var(Var),
	!,
	MessI = MessO,
	Args = Loc.
debug_call_location_exception_message( locate(cannot_locate(Spec,Ext)), _Loc, _MessI, Mess, Args ) :-
	Mess = 'Could not locate file specified by: ~w, and extensions: ~w',
	Args = [Spec,Ext].
debug_call_location_exception_message( Error, _Loc, _MessI, _Mess, _Args ) :-
	% fixme:
	throw( debug_call_caught(Error) ).

debug_mtx_dims( Rows, NRows, NCols ) :-
	length( Rows, NRows ),
	Rows = [Hdr|_],
	( is_list(Hdr) -> length(Hdr,NCols); functor(Hdr,_,NCols) ).

debug_message_prefixed( '', Standard, Standard ) :- !.
debug_message_prefixed( Pfx, Standard, Prefixed ) :-
	sub_atom( Standard, 0, 1, Aft, Fst ),
	downcase_atom( Fst, Low ),
	sub_atom( Standard, 1, Aft, 0, Right ),
	atomic_list_concat( [Pfx,' ',Low,Right], Prefixed ).

%% message_report( Format, Args, Kind ).
%
%  An Swi shortcut for printing messages.
%  The predicate first phrases onto a list the Format message
%  filled by Args, as it would do for debug( _, Format, Args ), 
%  then prints these lines as of Kind (error,warning,debug(_)).
%
%==
% ?- Mess = 'Destination:~w already pointed to:~w, repointing to:~w',
% |    F1 = 'file1', F2 = file2, F3 = file3,
% |    message_report( Mess, [F1,F2,F3], warning ).
%
% Warning: Destination:file1 already pointed to:file2, repointing to:file3
% 
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/02/28
%
message_report( Format, Args, Kind ) :-
	phrase('$messages':translate_message(debug(Format,Args)), Lines),
	print_message_lines(current_output, kind(Kind), Lines).



% datime_to_readable( -Ratom ).
% datime_to_readable( +Datime, -Ratom ).
%
% Ratom is a human readable representation of Datime.
% When Datime is missing the current datime is used.
% 
%==
% ?- datime_to_readable( Readable ).
%  Readable = 'At 15:13:36 on 2nd of Jul 2014'.
%==
%
% @author nicos angelopoulos
% @version  0.2 2014/7/2  Changed to date/9 and atom representation. Be ware if you are using 0.1
% @see debug_goal/3
% @tbd add precision for seconds.
%
datime_to_readable( Readable ) :-
	get_date_time( Current ),
	datime_to_readable( Current, Readable ).

datime_to_readable( date(Year,Month,Day,Hour,Mins,SecPrec,_,_,_), Readable ) :-
	three_letter_months( SymbolicMonths ),
	nth1( Month, SymbolicMonths, SymbMonth ),
	day_trailer( Day, Trailer ),
	Sec is floor( SecPrec ),
	Parts = ['At ',Hour,':',Mins,':',Sec,' on ',Day,Trailer,' of ',SymbMonth,' ',Year],
	atomic_list_concat( Parts, Readable ).

%% to_list( Term, Listed ).
%
% Wrap Term into a list of it not one already.
% For converting to lists see term_to_list/3.
%
%==
% ?- to_list( atom, List ).
% List = [atom].
%==
%
% @author nicos angelopoulos
% @version  0.2 2014/8/20
%
to_list( Either, List ) :-
	( (var(Either);(Either\=[_H|_T],Either\==[]) ) ->
		List = [Either]
		;
		List = Either
	).

% :- module( locate, [locate/3] ).


:- op( 400, fx, (/) ).  % bit uncomfortable about this. it is only used in 1 point... (see lib(os))

/** <module> locate

	This module is only here to prove a point- which I have now forgotten
*/

%% locate( +File, +Exts, -Locations ).
%
%  Find the exact Location of File that may have a number of extensions.
%  This should become the standard way to interface locating of reading in files.
%  =|Exts = any/*|=, is a special case where any file with matching extension
%  is returned. This case is slower than the rest.
%
%  As of 0.2 only existing files are located. Predicate throws error if file does not exist.
%
%==
%  locate( xyw, abc, Loc ).
%  ERROR: Unhandled exception: Cannot locate file with specification: xyw and extensions: abc
%  
%  
%==
% @author nicos angelopoulos
% @version  0.2 2014/4/24
%
locate( File, Exts, Location ) :-
	locate( File, Exts, '', Location),
	!.
locate( File, Exts, _Location) :-
	throw( locate(cannot_locate(File,Exts)) ).
	% print_message( error, locate(cannot_locate(File,Exts)) ).

locate( FileSpec, Exts, DefExt, Location ) :-
	( Exts == *; Exts == any ),
	!,
	expand_spec( FileSpec, Expanded ),
	locate_expanded( Expanded, [], '', AbsFile ),
	file_directory_name( AbsFile, Dir ),
	directory_contents( Dir, DFiles, _, _ ),
	directory_files( Dir, Entries ),
	include( exists_file, Entries, DFiles ),
	file_base_name( AbsFile, Bname ),
	include( loc_prefix_atom(Bname), DFiles, Candidates ),
	locate_in_prefix_candidates( Candidates, Bname, DefExt, LcBname ),
	directory_file_path( Dir, LcBname, Location ).

locate( File, Exts, DefExt, Location ) :-
	to_list( Exts, ExtsL ),
	expand_spec( File, Expanded ),
	locate_expanded( Expanded, ExtsL, DefExt, Location ).

% when file exists:
locate_expanded( Expanded, Exts, _DefExt, Location ) :-
	append( Exts, [''], Extensions ),
	Opts = [  extensions(Extensions), access(read),
	          file_errors(fail), solutions(first)   ], 
	absolute_file_name( Expanded, Location, Opts ).
/* commended out in version 0.2, fail with error if you cannot find it. this is for existing files.
locate_expanded( Expanded, _Exts, DefExt, Location ) :-
	Opts = [ access(none), solutions(all) ], 
	absolute_file_name( Expanded, Stem, Opts ),
	file_directory_name( Stem, Dir ),
	exists_directory( Dir ),
	!,
	file_name_extension( Stem, DefExt, Location ).
	*/

loc_prefix_atom( Pfx, Atom ) :-
	atom_concat( Pfx, _, Atom ).

% avoid adding the same extension twice
locate_in_prefix_candidates( [], File, DefExt, Location ) :-
	file_name_extension( _, DefExt, File ),
	!,
	Location = File.
locate_in_prefix_candidates( [], File, DefExt, Location ) :-
	!,
	file_name_extension( File, DefExt, Location ).
locate_in_prefix_candidates( [H|_T], File, _DefExt, Location ) :-
	file_name_extension( File, _Ext, H ),
	!,
	Location = H.
locate_in_prefix_candidates( [_H|T], File, DefExt, Location ) :-
	locate_in_prefix_candidates( T, File, DefExt, Location ).

expand_spec( File, Expanded ) :-
	atomic( File ),
	expand_file_name( File, [Expanded] ),
	!.
expand_spec( File, Expanded ) :-
	(File = _/_ ; File=  /_ ),
	!,
	term_atom( File, Atomic ),
	expand_file_name( Atomic, [Expanded] ).
expand_spec( Spec, Expanded ) :-
	% allow for alias(abc/def),
	arity( Spec, Alias, 1 ),
	!,
	arg( 1, Spec, TermOr ),
	expand_alias( Alias, Dir ),
	to_os_atom( TermOr, File ),
	directory_file_path( Dir, File, Expanded ).
expand_spec( Spec, Expanded ) :-
	to_os_atom( Spec, Expanded ),
	!.
expand_spec( File, File ).

expand_alias( Alias, Dir ) :-
	user:file_search_path( Alias, Dir ),
	atomic( Dir ),
	!.
expand_alias( Alias, Dir ) :-
	user:file_search_path( Alias, Expant ),
	expand_spec( Expant, Dir ).

to_os_atom( Term, File ) :-
	\+ var( Term ),
	to_os_atom_1( Term, File ).

to_os_atom_1( A/File, Path ) :-
	!,
	to_os_atom_1( A, Dir ),
	directory_file_path( Dir, File, Path ).
to_os_atom_1( File, File ).

:- multifile prolog:message//1.
:- discontiguous message//1.

% this is crucial for allowing cathcable exceptions that print when 
prolog:message(unhandled_exception(locate(Message))) -->
	{ debug( locate, 'Unhandled ~p', Message ) },
	message(Message).

prolog:message(locate(Message)) -->
	message(Message).

message( cannot_locate(File,Exts) ) -->
	[ 'Cannot locate file with specification: ~q and extensions: ~q'-[File,Exts] ].

%% get_date_time( -CurrDatime ).
%
% Just a wrapper to SWI's =|get_time(Stamp), stamp_date_time(Dtime).|=
% CurrDatime should be a date_time/1 term.
% SWI specific. Check YAP.
%
% @author nicos angelopoulos
% @version  0.1 2014/03/31 (some time well before)
%
get_date_time( Datime ) :-
     get_time( Stamp ),
     stamp_date_time( Stamp, Datime, local ).

% three_letter_months/1
% Unix (ls and ohters) compatible

three_letter_months( ['Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'] ).

day_trailer( 1, 'st' ) :- !.
day_trailer( 2, 'nd' ) :- !.
day_trailer( 3, 'rd' ) :- !.
day_trailer( _AllOther, 'th' ).

%% arity( ?Term, ?Name, ?Arity ).
%
%  This is the permissive version, if we detect atomic
%  we use functor/3 (the old stuff), otherwise we call 
%  compound_name_arity/3.
%  
arity( Term, Name, Arity ) :-
	current_predicate( compound_name_arity/3 ),
	\+ atomic( Term ),
	!,
	compound_name_arity( Term, Name, Arity ).
arity( Term, Name, Arity ) :-
	functor( Term, Name, Arity ).

